#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <QFileDialog>
#include <QSettings>
#include <QStandardPaths>
#include <QMessageBox>
#include <QRegExpValidator>
#include <QStringList>
#include "misc.h"
#include <stdio.h>
#include <QTimer>
#include <errno.h>
#include <string.h>
#include "zap.h"
#include <unistd.h>
#include "settingsdialog.h"
#include <QProcess>

//const char * version = u8"1.01";	// initial version that was functional enough to start collecting data.  some time around 6-15-2020
//const char * version = u8"1.1";	// don't respnd to comboBox change events while loading channels file,
												// added "scan" command line arg
//const char * version = u8"1.2";	// turn current row yellow when scanning
//const char * version = u8"1.3";	// Clean up text output
//const char * version = u8"1.4";	// Re-enable watch button when mplayer closed
//const char * version = u8"1.5";	// Allow specifying filename as 2nd arg when 1st arg is "scan".  12-7-2020
//const char * version = u8"1.6";	// Auto-detect number of adapters when set to 0 in settings, reverse usage order if < 0.  12-11-2020
//const char * version = u8"1.7";	// Added SNR graph and turn them both green when has_lock, handle CloseEvent, add 0 (adapter) first checkbox.  12-12-2020
//const char * version = u8"1.8";	// Log max_snr, min_snr, mean_snr and version.  12-13-2020
//const char * version = u8"1.9";	// Added program_initted bool, auto-generate logFname from fileName if !logFname.length.  12-18-2020
//const char * version = u8"1.10";	// Don't count snr's in first 4 seconds of zapping (like signal).  12-18-2020
//const char * version = u8"1.11";	// Log chip name and had_lock.  12-18-2020
//const char * version = u8"1.12";	// Wait up to 5 secs for zapper to stop.  12-18-2020
//const char * version = u8"1.13";	// Log lock_pct instead of 1/0.  12-20-2020
//const char * version = u8"1.14";	// Use %d to build watch command.  12-26-2020
//const char * version = u8"1.16";	// Add Recent Filenames and channels file on command line.  12-28-2020
//const char * version = u8"1.17";	// Add manufacturer to chip name.  1-1-2021
//const char * version = u8"1.18";	// Re-write command line arguments, handle -c (chip), -a (adapter) and -s (scan), remaining args must be filename.  1-1-2021
//const char * version = u8"1.19";	// Add -h for version and usage.  1-1-2021
//const char * version = u8"1.20";	// Add chip and adapter combo boxes.  1-1-2021
//const char * version = u8"1.21";	// Handle errors if cmdline adapter or chip not found.  1-8-2021
//const char * version = u8"1.22";	// Allow multiple watches, add %n (name) to watch command.  1-9-2021
//const char * version = u8"1.23";	// Add num_rdgs to the log.  1-10-2021
//const char * version = u8"1.24";	// Log current_adapter, start looking into /sys for chip info using QProcess.  1-15-2021
//const char * version = u8"1.25";	// Removed the unused Signal column from channels table (maybe some day).  1-16-2021
//const char * version = u8"1.26";	// Changed all perror() calls to fprintf(stderr).  1-29-2021
const char * version = u8"1.27";		// Honor numAdapters from .INI file, getp chip name from /tmp/dmesg.out if it exists.  1-31-2021


bool program_initted = false;

// parsed data from channels.conf file (100 channels max)
QString names[101];
QString channels[101];
unsigned int freqs[101];
unsigned short apids[101];
unsigned short vpids[101];
unsigned short sids[101];

// INI file settings
int numAdapters = 1;
int update_ms = 200;
bool do_zap = false;
bool scanning = false;
int monitorSeconds;
int logSeconds;
//bool useLastFirst = false;


// for logging elapsed time and channel info when logging monitor stats
time_t start_monitor;
time_t last_log_time = 0;
int scan_idx = -1;

/* Tuner parameters
int adapter = 0, frontend = 0, demux = 0, dvr = 0;
int vpid, apid, sid, pmtpid = 0;
int pat_fd, pmt_fd;
int frontend_fd, audio_fd, video_fd;
*/
char chip_name[1024];
int specific_adapter = -1;
char specific_chip[128];
int current_adapter;

MainWindow::MainWindow(QWidget *parent)
	: QMainWindow(parent)
	, ui(new Ui::MainWindow)
{
	specific_adapter = -1;
	memset(specific_chip, 0, 128);

	ui->setupUi(this);

	newfile = "";
	save_filename = true; // only false if arg[1] is "-s" (scan)
	loading_file = false;
	current_channel_idx = 0;
	pZapper = nullptr;
	pValidator = nullptr;
	pWatcher = nullptr;
	num_rdgs = 0;

	// Read window size/position from .INI file
	iniFname = QStandardPaths::locate(QStandardPaths::HomeLocation, QString(), QStandardPaths::LocateDirectory)
	 + ".strength.ini";
	QSettings settingsFile(iniFname, QSettings::NativeFormat);

	int xspot = settingsFile.value("xpos", 0).toInt();
	int yspot = settingsFile.value("ypos", 0).toInt();
	move(xspot, yspot);
	fileName = settingsFile.value("channels_file", "").toString();
	logFname = settingsFile.value("log_filename", "").toString();
	numAdapters = settingsFile.value("number_of_adapters", 1).toInt();
	update_ms = settingsFile.value("update_ms", 200).toInt();
	monitorSeconds = settingsFile.value("monitor_seconds", 30).toInt();
	logSeconds = settingsFile.value("log_seconds", 10).toInt();
	watchCommand = settingsFile.value("watch_command", "mplayer -cache 1024").toString();
	numRecentFiles = settingsFile.value("MaxRecentFiles", 10).toInt();
	if (numRecentFiles > MAX_RECENT_FILES)
		numRecentFiles = MAX_RECENT_FILES;

	setWindowTitle("Signal Strength V" + QString(version) + " - No channels loaded");

	addAction(ui->actionScanSelected);
	addAction(ui->actionStop_Monitoring);
	addAction(ui->actionSettings);
	addAction(ui->actionLoad);
	addAction(ui->actionExit);

	ui->menuRecent_Channels_Files->setTitle(QCoreApplication::translate("MainWindow", "Recent Files", nullptr));
	for (int i = 0; i < MAX_RECENT_FILES; ++i) {
		recentFileActs[i] = new QAction(this);
		recentFileActs[i]->setVisible(false);
		connect(recentFileActs[i], SIGNAL(triggered()), this, SLOT(openRecentFile()));
		ui->menuRecent_Channels_Files->addAction(recentFileActs[i]);
	}

	updateRecentFileActions(settingsFile);

	connect(this, SIGNAL(sendWatchButtonClick()), this, SLOT(on_watchButton_clicked()), Qt::QueuedConnection);

	ui->channelTable->setSelectionBehavior(QAbstractItemView::SelectRows);
	QStringList colHeadings = { "Idx", "Chan", "Name", "Mhz", "APID", "VPID", "SID" };
//	QStringList colHeadings = { "Idx", "Chan", "Name", "Mhz", "Signal", "APID", "VPID", "SID" };
	ui->channelTable->setColumnCount(colHeadings.count());
	ui->channelTable->setHorizontalHeaderLabels(colHeadings);

	for (int c = 0; c < ui->channelTable->columnCount(); c++) {
		int w = settingsFile.value("col_width_" + QString::number(c), 70).toInt();
		ui->channelTable->setColumnWidth(c, w);
	}
} // MainWindow()

MainWindow::~MainWindow()
{
	if (pWatcher)
		pWatcher->kill();

	pWatcher = nullptr;

	if (pZapper) {
		if (pZapper->zapping) {
			do_zap = false;
			MySleep(update_ms * 2); // give pZapper time to emit its done finished and get deleted in the zapperFinished() slot
		}
		if (pZapper) {  // just in case it didn't work...
			delete pZapper;
			pZapper = nullptr;
		}
	}

	if (pValidator)
		delete pValidator;
	pValidator = nullptr;

	QSettings settingsFile(iniFname, QSettings::NativeFormat);
	settingsFile.setValue("xpos", pos().x());
	settingsFile.setValue("ypos", pos().y());
	if (save_filename)
		settingsFile.setValue("channels_file", fileName);
	settingsFile.setValue("log_seconds", ui->secsEdit->text());

	for (int c = 0; c < ui->channelTable->columnCount(); c++) {
		int w = ui->channelTable->columnWidth(c);
		settingsFile.setValue("col_width_" + QString::number(c) , w);
	}

	for (int i = 0; i < numRecentFiles; ++i) {
		if (recentFileActs[i])
			delete recentFileActs[i];
	}

	delete ui;
}// ~MainWindow()

void MainWindow::showEvent(QShowEvent *)
{
	QTimer::singleShot(50, this, SLOT(on_init_program()));
}

void MainWindow::on_init_program(void)
{
	if (program_initted)
		return;
	program_initted = true;
//	was_watching = false;

	ui->chip_comboBox->addItem("Any Chip");
	ui->adapter_comboBox->addItem("Any Adapter");

	int num_adapters;
	FILE * cmd = popen("/bin/ls /dev/dvb | wc -l", "r");
	fscanf(cmd, "%d", &num_adapters);
	pclose(cmd);

	QStringList chipNames;

	for (int adapter = 0; adapter < num_adapters; adapter ++) {
		ui->adapter_comboBox->addItem("Adapter " + QString::number(adapter));

		// Get the adapter's chip name
		// dmesg | grep 'registering adapter'    
		// [   16.797196] cx23885 0000:0b:00.0: DVB: registering adapter 0 frontend 0 (Samsung S5H1411 QAM/8VSB Frontend)...
		// [   16.849681] cx23885 0000:0b:00.0: DVB: registering adapter 1 frontend 0 (Samsung S5H1411 QAM/8VSB Frontend)...
		// [281448.513000] em28xx 3-1:1.0: DVB: registering adapter 2 frontend 0 (LG Electronics LGDT3306A VSB/QAM Frontend)...
		// [281448.522076] em28xx 3-1:1.0: DVB: registering adapter 3 frontend 0 (LG Electronics LGDT3306A VSB/QAM Frontend)...
		char chpName[1024];
		memset(chpName, 0, 1024);
		char dmsg_cmd[2048];
		sprintf(dmsg_cmd, "dmesg | sed -n -r 's/\\[[ 0-9.]+\\]\\s+(\\S+)\\s+\\S+\\s+DVB: registering adapter %d frontend [0-9] \\((\\S+) .+/\\1_\\2/p'", adapter);
	
		FILE * dmesg = popen(dmsg_cmd, "r");
		if (dmesg) {
			fgets(chpName, 1023, dmesg);
			pclose(dmesg);
		}
		
// The following is being developed to extract car/chip info for each adapter from /sys info and lspci since dmesg is only good for about 4 days
/*
		QProcess p;
		p.start("/bin/bash");
		p.waitForStarted();
		char cmd[256];
		snprintf(cmd, 255, "cat /sys/class/dvb/dvb0.frontend%d/device/subsystem_device\n", adapter);
		p.write(cmd);
		p.waitForReadyRead(500);
		QString device = p.readAll().trimmed();
		snprintf(cmd, 255, "cat /sys/class/dvb/dvb%d.frontend0/device/subsystem_vendor\n", adapter);
		p.write(cmd);
		p.waitForReadyRead(500);
		QString tuner_device = p.readAll().trimmed();
		p.write("/usr/sbin/lspci -v");
		QString mfg;
		QString model;
		QString chipName;
		QString idLine;
		QRegExp re("\\d+");  // a digit (\d+), one or more times (*)

		bool readable = p.waitForReadyRead(1500);
		while (readable) {
			char buf[1024];
			p.readLine(buf, 1024);
			if (!isspace(buf[0])) {  // found device ID line
				idLine = QString(buf);
					readable = p.waitForReadyRead(100);

		p.write("exit\n");
		p.waitForFinished();
*/
/* lspci -v
0e:00.0 Multimedia video controller: Conexant Systems, Inc. CX23885 PCI Video and Audio Decoder (rev 02)
	Subsystem: DViCO Corporation Device d618
	Flags: bus master, fast devsel, latency 0, IRQ 18
	Memory at f7400000 (64-bit, non-prefetchable) [size=2M]
	Capabilities: <access denied>
	Kernel driver in use: cx23885
	Kernel modules: cx23885

04:00.0 Multimedia video controller: Conexant Systems, Inc. CX23887/8 PCIe Broadcast Audio and Video Decoder with 3D Comb (rev 04)
	Subsystem: Hauppauge computer works Inc. Device 6b18
	Flags: bus master, fast devsel, latency 0, IRQ 18
	Memory at f7000000 (64-bit, non-prefetchable) [size=2M]
	Capabilities: <access denied>
	Kernel driver in use: cx23885
	Kernel modules: cx23885
*/

		int cnlen = strlen(chpName);

		// See if there is a saved copy of dmesg that contains the chip
		if (cnlen < 3) {
			fprintf(stderr, "Checking /tmp/dmesg.out for chip name\n");
			sprintf(dmsg_cmd, "cat /tmp/dmesg.out | sed -n -r 's/\\[[ 0-9.]+\\]\\s+(\\S+)\\s+\\S+\\s+DVB: registering adapter %d frontend [0-9] \\((\\S+) .+/\\1_\\2/p'", adapter);
			FILE * tmpdmesg = popen(dmsg_cmd, "r");
			if (tmpdmesg) {
				fgets(chpName, 1023, dmesg);
				pclose(tmpdmesg);
				cnlen = strlen(chpName);
			}
		}

		if (cnlen > 1) {
			chpName[cnlen-1] = '\0'; // remove trailing \n
			QString chip(chpName);
			chipNames.append(chip);
		}
	}
	chipNames.removeDuplicates();
	ui->chip_comboBox->addItems(chipNames);

	pValidator = new QRegExpValidator(QRegExp("\\d*"), this);
	ui->secsEdit->setValidator(pValidator);
	ui->secsEdit->setText(QString::number(logSeconds));

	if (fileName.length()) {
		save_filename = false; // don't save the previous filename as most recently used
		load_channels();
		save_filename = true;
	} // fileName.length()

	updateStrength(0.0, 0.0, false);
	timer.setInterval(monitorSeconds * 1000);
	timer.stop();
	connect(&timer, SIGNAL(timeout()), this, SLOT(on_actionScanSelected_triggered())); //, Qt::QueuedConnection);
	timer.setInterval((monitorSeconds + 4) * 1000);
	timer.setSingleShot(true);
	scanning = false;
	scan_idx = -1;
	num_rdgs = 0;
	ui->channel_comboBox->setCurrentIndex(0);
	MySleep(3 * update_ms);
	
	//ui->textEdit->append(QCoreApplication::arguments().at(0)); // executable filename
	//for (int a = 1; a < QCoreApplication::arguments().count(); a++)
	//	ui->textEdit->append(QCoreApplication::arguments().at(a)); // first argument

	QStringList appArgs = QCoreApplication::arguments();
	
	// if there are command line arguments...
	if (appArgs.count() > 1) {  // more than just the executable name...

		// check for -a (adapter) switch
		if (appArgs.contains("-a")) {

			int argnum = appArgs.indexOf("-a");

			int ac = appArgs.count();
			if (ac < argnum + 2) {
				fprintf(stderr, "No adapter specified after -a\n");
				close();
				return;
			}

			QString a = appArgs[argnum+1];
			QRegExp re("\\d+");  // a digit (\d+), one or more times (*)

			if (!re.exactMatch(a)) {
				fprintf(stderr, "Adapter must be a number\n");
				close();
				return;
			}

			int requestedAdapter = a.toInt();
			if ((requestedAdapter < 0) || (requestedAdapter > num_adapters - 1)) {
				ui->textEdit->append("Invalid adapter: " + a);
				char errbuf[128];
				snprintf(errbuf, 127, "invalid adapter: %d\n", requestedAdapter);
				fprintf(stderr, "%s", errbuf); 
				specific_adapter = requestedAdapter;
			}
			else
				ui->adapter_comboBox->setCurrentIndex(requestedAdapter + 1);

			appArgs.removeAt(argnum+1);
			appArgs.removeAt(argnum);
		}

		// check for -c (chip) switch
		if (appArgs.contains("-c")) {

			int argnum = appArgs.indexOf("-c");

			int ac = appArgs.count();
			if (ac < argnum + 2) {
				fprintf(stderr, "No chip specified after -c\n");
				close();
				return;
			}

			QString c = appArgs[argnum+1];

			if (!c.length()) {
				fprintf(stderr, "No chip specified after -c\n");
				close();
				return;
			}

			int chipIdx = ui->chip_comboBox->findText(c);
			if (chipIdx >= 0)
				ui->chip_comboBox->setCurrentIndex(chipIdx);
			else {
				ui->textEdit->append("Chip not found in dmesg: " + c);
				strncpy(specific_chip, QS_to_char(c), 128);
				char errbuf[128];
				snprintf(errbuf, 127, "Chip not found in dmesg: %s\n", specific_chip);
				fprintf(stderr, "%s", errbuf); 
			}

			appArgs.removeAt(argnum+1);
			appArgs.removeAt(argnum);
		}

		// check for -s (scan) switch
		if (appArgs.contains("-s")) {
			save_filename = false;

			int argnum = appArgs.indexOf("-s");
			int ac = appArgs.count();

			if (ac > argnum + 1) {
				fileName = appArgs.at(argnum + 1);
				load_channels();
			}

			ui->channelTable->selectAll();
			on_actionScanSelected_triggered();
			MySleep(3000);
			while (scanning)
				MySleep(1000);
			close();
			return;
		}
	} // QCoreApplication::arguments().count() > 1

	if (appArgs.count() > 1) {
		QFileInfo check_file(appArgs.at(1));
		if (check_file.exists() && check_file.isReadable()) {
			fileName = appArgs.at(1);
			load_channels();
		}
		else
			ui->textEdit->append("Unknown command line argument: " + appArgs.at(1));
	} // QCoreApplication::arguments().count() > 1

	QString winTitle = "Signal Strength V" + QString(version)
	 + " - " + QString::number(ui->channel_comboBox->count()-1)
	  + " channels loaded from " + QFileInfo(fileName).fileName();

	if (specific_adapter >= 0)
		winTitle += ", Using adapter " + QString::number(specific_adapter);
	if (strlen(specific_chip))
		winTitle += ", Using adapter chip " + QString(specific_chip);

	setWindowTitle(winTitle);

} // on_init_program()

void MainWindow::on_actionExit_triggered()
{
	close();
}

void MainWindow::on_actionLoad_triggered()
{
	QString homeDir = QStandardPaths::locate(QStandardPaths::HomeLocation, QString(), QStandardPaths::LocateDirectory);
	if (!newfile.length())
		newfile = QFileDialog::getOpenFileName(this, ("Open File"), homeDir, ("Channels (*.conf)"));
	if (!newfile.length())
		return;

	// make sure the file exists
	QFileInfo check_file(newfile);
	if (!check_file.exists() || !check_file.isReadable()) {
		QString errstr = "Could not read from " + newfile;
		QMessageBox msgBox;
		msgBox.setText("Error!");
		msgBox.setInformativeText(errstr);
		msgBox.setStandardButtons(QMessageBox::Ok); // | QMessageBox::Discard | QMessageBox::Cancel);
		//msgBox.setDefaultButton(QMessageBox::Save);
		msgBox.exec();
		newfile = "";
		return;
	}

	fileName = newfile;
	newfile = "";
	load_channels();
}

void MainWindow::openRecentFile()
{
	QAction *action = qobject_cast<QAction *>(sender());
	if (action) {
		newfile = action->data().toString();
		on_actionLoad_triggered();
	}
}

void MainWindow::load_channels(void)
{

	FILE * cf = fopen(QS_to_char(fileName), "r");
	if (cf) {

		loading_file = true;

		ui->channel_comboBox->clear();
		//ui->textEdit->setText("");
		ui->channelTable->setRowCount(0);
		ui->channel_comboBox->addItem("Select Channel");

		int idx = 0;
		char buf[1024];

		while ((idx < 100) && (fgets(buf, 1023, cf))) {

			// KCTS-HD:189028615:8VSB:49:53:3

			char * cptr = strchr(buf, ':');
			*cptr = '\0'; // terminate name
			names[idx] = QString(buf);

			char * next = cptr + 1;
			cptr = strchr(next, ':'); // find end of frequency
			*cptr = '\0'; // terminate frequency
			freqs[idx] = (unsigned int) atoi(next);

			cptr++; // point at 8VSB
			cptr = strchr(cptr, ':'); // find end of 8VSB
			next = cptr + 1; // point at APID
			cptr = strchr(next, ':'); // find end of APID
			*cptr = '\0'; // terminate APID
			apids[idx] = (unsigned short) atoi(next);

			next = cptr + 1; // point at VPID
			cptr = strchr(next, ':'); // find end of VPID
			*cptr = '\0'; // terminate VPID
			vpids[idx] = (unsigned short) atoi(next);

			next = cptr + 1; // point at transport
			sids[idx] = (unsigned short) atoi(next);

			int baseFreq = 177028615; // channel 7's center freq
			int baseChan = 7;
			if (freqs[idx] > 3e8) {
				baseFreq = 473028615; // channel 14's center freq
				baseChan = 14;
			}
			int chanNum = baseChan + ((freqs[idx] - baseFreq) / 6000000);
			channels[idx] = QString::number(chanNum) + "." + QString::number(sids[idx]);

			ui->channel_comboBox->addItem(names[idx]);

			int col_num = 0;

			ui->channelTable->insertRow(idx);
			// colHeadings = { "Idx", "Chan", "Name", "Mhz", "Signal", "APID", "VPID", "SID" };
			QTableWidgetItem * newItem = new QTableWidgetItem(idx); // channel number (VHF/UHF channel.SID
			newItem->setData(Qt::EditRole, idx + 1);
			newItem->setTextAlignment(Qt::AlignCenter);
			ui->channelTable->setItem(idx, col_num++, newItem);  // row, col, newItem

			newItem = new QTableWidgetItem(channels[idx], idx); // Channum (7.1, 5.3)
			newItem->setTextAlignment(Qt::AlignCenter);
			ui->channelTable->setItem(idx, col_num++, newItem);

			newItem = new QTableWidgetItem(names[idx], idx); // name (callsign)
			newItem->setTextAlignment(Qt::AlignCenter);
			ui->channelTable->setItem(idx, col_num++, newItem);

			newItem = new QTableWidgetItem(idx); // Mhz
			newItem->setData(Qt::EditRole, freqs[idx]/1000000);
			newItem->setTextAlignment(Qt::AlignCenter);
			ui->channelTable->setItem(idx, col_num++, newItem);

/*
			newItem = new QTableWidgetItem(idx); // Signal strength (zero, now, will be filled in after zap)
			newItem->setData(Qt::EditRole, 0);
			newItem->setTextAlignment(Qt::AlignCenter);
			ui->channelTable->setItem(idx, col_num++, newItem);
*/
			newItem = new QTableWidgetItem(idx); // APID
			newItem->setData(Qt::EditRole, apids[idx]);
			newItem->setTextAlignment(Qt::AlignCenter);
			ui->channelTable->setItem(idx, col_num++, newItem);
			
			newItem = new QTableWidgetItem(idx); // VPID
			newItem->setData(Qt::EditRole, vpids[idx]);
			newItem->setTextAlignment(Qt::AlignCenter);
			ui->channelTable->setItem(idx, col_num++, newItem);
			
			newItem = new QTableWidgetItem(idx); // SID
			newItem->setData(Qt::EditRole, sids[idx]);
			newItem->setTextAlignment(Qt::AlignCenter);
			ui->channelTable->setItem(idx, col_num++, newItem);

			idx++;
		}
		fclose(cf);

		ui->channel_comboBox->setEnabled(true);

		QString winTitle = "Signal Strength V" + QString(version) + " - " + QString::number(idx) + " channels loaded from " + QFileInfo(fileName).fileName();

		if (specific_adapter >= 0)
			winTitle += ", Using adapter " + QString::number(specific_adapter);
		if (strlen(specific_chip))
			winTitle += ", Using adapter chip " + QString(specific_chip);

		setWindowTitle(winTitle);

		if (save_filename)
			saveRecentFilenames();

		loading_file = false;
	} // if cf
	else {
		QString errstr = "Could not open " + fileName + ": " + QString(strerror(errno));
		QMessageBox msgBox;
		msgBox.setText("Error!");
		msgBox.setInformativeText(errstr);
		msgBox.setStandardButtons(QMessageBox::Ok); // | QMessageBox::Discard | QMessageBox::Cancel);
		//msgBox.setDefaultButton(QMessageBox::Save);
		msgBox.exec();
	}
} // load_channels(void)

void MainWindow::on_channel_comboBox_currentIndexChanged(int index)
{
	if (loading_file)
		return;

	current_channel_idx = index - 1;
	logSeconds = ui->secsEdit->text().toInt();

	if (index == 0)
	{
/*		if (pWatcher) {
			//was_watching = true;
			pWatcher->kill();
			pWatcher = nullptr;
		} */
//		else
//			was_watching = false;

		do_zap = false;
	 	//ui->textEdit->setText("");
		MySleep(update_ms * 2);
	 	if (pZapper) {
			MySleep(1000);
			if (pZapper) {
				delete pZapper;
				pZapper = nullptr;
			}
		}
		updateStrength(0.0, 0, false);
		return;
	}

	//if (!scanning)
	//	ui->textEdit->setText("");

	ui->textEdit->append("Name:  " + ui->channel_comboBox->currentText());
	ui->textEdit->append("Channel:  " + channels[index-1]
	 + "\nFrequency:  " + QString::number(((float)freqs[current_channel_idx])/1e6, 'f', 3)); // + " Mhz\nAudio PID:  "
	 //+ QString::number(apids[current_channel_idx]) + "\nVideo PID:  " + QString::number(vpids[current_channel_idx]) + "\nStream ID:  " + QString::number(sids[current_channel_idx]));

	updateStrength(0.0, 0, false);

	// Create an instance of Zapper that runs a background thread to monitor channel strength, emitting newReading signales every update_ms mS
	pZapper = new Zapper(freqs[current_channel_idx], apids[current_channel_idx], vpids[current_channel_idx], sids[current_channel_idx]);

	// Connect the signals and slots
	connect(pZapper, SIGNAL(newReading(float, unsigned int, bool)), this, SLOT(updateStrength(float, unsigned int, bool)), Qt::QueuedConnection);
	connect(pZapper, SIGNAL(sendInfo(const char*)), this, SLOT(showInfo(const char*)), Qt::DirectConnection); // QueuedConnection);

	// Setup callback for cleanup when it finishes
	connect(pZapper, SIGNAL(finished()), this, SLOT(zapperFinished()));
	
	if (pZapper->open_frontend() || (!pZapper->ready)) {
		delete pZapper;
		pZapper = nullptr;
		ui->textEdit->append("\nFailed to open frontend");
		scanning = false;
		return;
	}

	// disable/enable widgets
	ui->channel_comboBox->setEnabled(false);
	ui->stopButton->setEnabled(true);
	ui->watchButton->setEnabled(!scanning);
	ui->actionLoad->setEnabled(false);
	ui->actionStop_Monitoring->setEnabled(true);

	ui->textEdit->append("Channel tuned, monitoring..."); // Ready to start monitoring\n");

	// Run the thread
	do_zap = true; // when this gets set false, the Thermometer thread about to be created will exit
	pZapper->start(); // This invokes Zapper::run in a new thread
} // on_channel_comboBox_currentIndexChanged()

void MainWindow::zapperFinished(void)
{
	//if (pWatcher) {
	//	pWatcher->kill();
		//delete pWatcher;  // handled in on_watchFinished()
		//pWatcher = nullptr;
	//}

	if (!pZapper)
		return;
	delete pZapper;
	pZapper = nullptr;

	updateStrength(0.0, 0, false);
	ui->textEdit->append("Monitoring stopped");

	ui->channel_comboBox->setEnabled(true);
	ui->stopButton->setEnabled(false);
	ui->watchButton->setEnabled(false);
	ui->actionLoad->setEnabled(true);
	ui->actionStop_Monitoring->setEnabled(false);
}

// receives signal sendInfo(const char* info);
void MainWindow::showInfo(const char * txt)
{
	ui->textEdit->append(QString(txt));
	MySleep(10);
} // showInfo()

// receives signal newReading(float pct, bool has_lock);
void MainWindow::updateStrength(float pct, unsigned int snr, bool has_lock)
{
	// barFrame is located at 40,30 and is 120 wide by 380 high.  It has a 2px wide frame
	// bar is located inside barFrame at 2, 376-barHeight and is 116 wide by 376 at full scale

	if (pct > 1e2)
		pct = 1e2;
	if (pct < 0)
		pct = 0.0;

	// don't count readings that are less than 4 seconds since zapping first started
	if (scanning && (time(NULL) - start_monitor > 4)) {

		lock_count += has_lock ? 1 : 0;

		mean_snr += snr;
		if (snr > max_snr)
			max_snr = snr;
		if (snr < min_snr)
			min_snr = snr;

		mean += pct;
		num_rdgs++;
		if (pct > max)
			max = pct;
		if (pct < min)
			min = pct;

	}

	if (!scanning && pZapper && do_zap && ui->logCheckBox->isChecked() && (time(NULL) - last_log_time >= logSeconds)) {
		last_log_time = time(NULL);
		FILE * f = fopen(QS_to_char(logFname), u8"a");
		if (!f) {
			ui->logCheckBox->setChecked(false);
			char ebuf[4096];
			snprintf(ebuf, 4095, "Error opening log file %s!  %s\n", QS_to_char(logFname), strerror(errno));
			ui->textEdit->append(QString(ebuf));
		}
		else {
			//Chan_List#, channel, Mhz, max, min, mean, max_snr, min_snr, mean_snr, has_lock, num_rdgs, secs_since_start, adapter, name, chip_name, version, timestamp, time_date
			// idx channum freq max min mean max_snr min_snr mean_snr has_lock num_rdgs elapsed_secs current_adapter name chip_name version dt ctime

			char nameBuf[32];
			char chanBuf[32];
			strncpy(nameBuf, QS_to_char(names[current_channel_idx]), 32);
			strncpy(chanBuf, QS_to_char(channels[current_channel_idx]), 32);
			fprintf(f, "%d, %s, %d, 0.0, 0.0, %1.1f, 0, 0, %d, %1.1lf, %d, 0, %d, %s, %s, %s, %ld, %s",
			 current_channel_idx + 1,
		 	 chanBuf,
			 freqs[current_channel_idx] / 1000000,
			 has_lock ? pct : pct * -1.0,
			 has_lock ? snr : snr * -1,
			 has_lock ? 1e2 : 0.0,
			 1,
			 current_adapter,
			 nameBuf,
			 chip_name,
			 version,
			 last_log_time, ctime(&last_log_time)
			);
			fclose(f);
		}
	}
		
	ui->pctLabel->setText(QString::number((int)(pct+0.5)) + "%");
	ui->snrLabel->setText(QString::number(snr));

	int barHeight = (int) ((double) (ui->barFrame->height()-4) * ((pct + 0.5) / 1e2))-1;
	if (barHeight < 1)
		barHeight = 1;
	int barTop = (ui->barFrame->height() - barHeight) - 2;

	if (has_lock) {
		ui->bar->setStyleSheet("QFrame { background-color: #6c8; }");
		ui->snrBar->setStyleSheet("QFrame { background-color: #6c8; }");
	}
	else {
		ui->bar->setStyleSheet("QFrame { background-color: #9af; }");
		ui->snrBar->setStyleSheet("QFrame { background-color: #9af; }");
	}

	ui->bar->resize(ui->barFrame->width()-4, barHeight);
	ui->bar->move(2, barTop);

	// use 300 as 100% SNR
	double snrPct = snr/3e2;
	if (snrPct > 1.0)
		snrPct = 1.0;
	if (snrPct < 0.0)
		snrPct = 0.0;
	barHeight = (int) ((double) (ui->snrFrame->height()-4) * snrPct);
	if (barHeight < 1)
		barHeight = 1;
	barTop = (ui->snrFrame->height() - barHeight)-2;

	ui->snrBar->resize(ui->snrFrame->width()-4, barHeight);
	ui->snrBar->move(2, barTop);

	ui->hasLock->setVisible(has_lock);

	if (has_lock && pWatcher && ui->watchButton->isEnabled()) {
		pWatcher = nullptr;
		emit sendWatchButtonClick();
	}

	//MySleep(10);

} // updateStrength()

void MainWindow::on_actionStop_Monitoring_triggered()
{
	timer.stop();
	do_zap = false;
	scanning = false;
	scan_idx = -1;
	num_rdgs = 0;
	ui->channel_comboBox->setCurrentIndex(0);
	MySleep(3 * update_ms);
	if (pWatcher) {
		pWatcher->kill();
		pWatcher = nullptr;
	}
}

void MainWindow::on_stopButton_clicked()
{
	on_actionStop_Monitoring_triggered();
}

// Open the settings window
void MainWindow::on_actionSettings_triggered()
{
	settingsDialog sd(this, logFname, watchCommand);
	connect(&sd, SIGNAL(sendInfo(const char*)), this, SLOT(showInfo(const char*)), Qt::DirectConnection); // QueuedConnection
	connect(&sd, SIGNAL(sendSetting(QString, QString, bool)), this, SLOT(saveSetting(QString, QString, bool)), Qt::QueuedConnection);
	sd.exec();
}

void MainWindow::saveSetting(QString settingName, QString settingValue, bool saveIt)
{
	// The other setting are global and are applied in the settingsDialog
	if (settingName == "log_filename")
		logFname = settingValue;
	else if (settingName == "watch_command")
		watchCommand = settingValue;

	if (saveIt) {
		QSettings settingsFile(iniFname, QSettings::NativeFormat);
		settingsFile.setValue(settingName, settingValue);
	}
}

void MainWindow::on_channelTable_cellDoubleClicked(int row, int column)
{
	timer.stop();
	num_rdgs = 0;
	scanning = false;
	scan_idx = -1;

	//if (pWatcher)
	//	was_watching = true;

	if (do_zap && pZapper) {
		do_zap = false;
		MySleep(update_ms * 2);
		if (pZapper)
			return;
	}

	QTableWidgetItem * pCell = ui->channelTable->item(row, column);
	ui->channel_comboBox->setCurrentIndex(pCell->type()+1);
}

// Scan through all selected rows' channels from the channels table and
// monitor them for monitorSeconds, logging max, min, mean, max_snr, min_snr and mean_snr in logFname
void MainWindow::on_actionScanSelected_triggered()
{
	// if no log filename specified, use the channels.conf file with .conf changed to .dat or add .dat
	if (!logFname.length()) {
		if (fileName.endsWith(".conf", Qt::CaseInsensitive)) {
			logFname = fileName;
			logFname.replace(".conf", ".dat", Qt::CaseInsensitive);
		}
		else
			logFname = fileName + ".dat";
	}
	
	timer.stop(); // probably not necessary
	if (!scanning)
		ui->textEdit->setText("");

	// just finished monitoring a channel, log stats
	if (num_rdgs > 0) {
		mean_snr /= num_rdgs;
		mean /= (float) num_rdgs;
		double lock_pct = ((double) lock_count / (double) num_rdgs) * 1e2; 

		FILE * f = fopen(QS_to_char(logFname), u8"a");
		if (!f) {
			scanning = false;
			char ebuf[4096];
			snprintf(ebuf, 4095, "Error opening log file %s!  %s\n", QS_to_char(logFname), strerror(errno));
			ui->textEdit->append(QString(ebuf));
			return;
		}
		time_t dt = time(NULL);
		char nameBuf[32];
		char chanBuf[32];
		strncpy(nameBuf, QS_to_char(names[scan_idx]), 32);
		strncpy(chanBuf, QS_to_char(channels[scan_idx]), 32);

		//Chan_List#, channel, Mhz, max, min, mean, max_snr, min_snr, mean_snr, has_lock, num_rdgs, secs_since_start, adapter, name, chip_name, version, timestamp, time_date
		// idx channum freq max min mean max_snr min_snr mean_snr lock_pct num_rdgs elapsed_secs name current_adapter chip_name version dt ctime

		fprintf(f, "%d, %s, %d, %1.1f, %1.1f, %1.1f, %d, %d, %d, %1.1lf, %d, %ld, %d, %s, %s, %s, %ld, %s",
		 scan_idx + 1,
		 chanBuf,
		 freqs[scan_idx] / 1000000,
		 max, min, mean,
		 max_snr, min_snr, mean_snr,
		 lock_pct,
		 num_rdgs,
		 dt - start_monitor,
		 current_adapter,
		 nameBuf,
		 chip_name,
		 version,
		 dt, ctime(&dt)
		);
		fclose(f);
	} // num_rdgs > 0

	// de-highlight the row
	QTableWidgetItem * pCurrentItem = nullptr;

	for (int yr = 0; yr < ui->channel_comboBox->count()-1; yr++) {
		pCurrentItem = ui->channelTable->item(yr, 1);
		if (scan_idx == pCurrentItem->type()) {
			for (int c = 0; c < ui->channelTable->columnCount(); c++)
				ui->channelTable->item(yr, c)->setBackground(Qt::white);
			MySleep(5);
			break;
		}
	}

	// stop the zapper
	if (do_zap && pZapper) {
		do_zap = false;

		double stop_zap_time = double_prec_time();
		while (pZapper && (double_prec_time() - stop_zap_time < 5.0))
			MySleep(2 * update_ms); // give the zapper time to stop and be deleted in zapperFinished()
		
		// something went wrong, stop scanning
		if (pZapper) {
			ui->textEdit->append("Zapper failed to stop, aborting scan!");
			scanning = false;
			return;
		}
		ui->textEdit->append("");
	}

	// no more rows selected
	if (!ui->channelTable->selectionModel()->selectedRows().count()) {
		scanning = false;
		scan_idx = -1;
		num_rdgs = 0;
		//ui->channel_comboBox->setCurrentIndex(0);
		ui->textEdit->append("Scan complete");
		return;
	}

	// init stats for channel about to be zapped
	num_rdgs = 0;
	max = -3e6;
	min = 3e6;
	mean = 0.0;
	max_snr = 0;
	min_snr = 10000;
	mean_snr = 0;
	lock_count = 0;

	scanning = true;

	// get the next selection from the table
	int r;
	
	QTableWidgetItem * pItem = nullptr;

	for (r = 0; r < ui->channel_comboBox->count()-1; r++) {
		pItem = ui->channelTable->item(r, 1);
		if (pItem && pItem->isSelected()) {
			for (int c = 0; c < ui->channelTable->columnCount(); c++) {
				ui->channelTable->item(r, c)->setSelected(false);
				ui->channelTable->item(r, c)->setBackground(Qt::yellow); //lightGray);
			}
			MySleep(5);
			break;
		}
		else
			pItem = nullptr;
	}

	// if no more selections, stop (should never happen since selected row count checked above)
	if (!pItem) {
		scanning = false;
		scan_idx = -1;
		num_rdgs = 0;
		//ui->channel_comboBox->setCurrentIndex(0);
		ui->textEdit->append("Scan completed");
		//MySleep(3 * update_ms);
		return;
	}

	// save the index
	scan_idx = pItem->type();
	
	// set the timer to come back here in monitorSeconds + 4 (first seconds are for channel to get tuned and are not counted)
	timer.start();

	//ui->textEdit->append("Tuning channel " + QString::number(scan_idx + 1) + " - " + names[scan_idx]);
	ui->textEdit->append("Tuning channel " + channels[scan_idx] + " - " + names[scan_idx]);

	start_monitor = time(NULL);

	// set the index of the comboBox to start
	ui->channel_comboBox->setCurrentIndex(scan_idx + 1);
}

void MainWindow::on_logCheckBox_clicked()
{
	logSeconds = ui->secsEdit->text().toInt();
}

void MainWindow::on_channelTable_itemSelectionChanged()
{
	ui->actionScanSelected->setEnabled(ui->channelTable->selectedItems().count() > 0);
}

void MainWindow::on_watchButton_clicked()
{
	if (pZapper && pZapper->zapping) {  // should not be here if not true
		char devName[64];

		snprintf(devName, 64, "/dev/dvb/adapter%d/dvr0", pZapper->currentAdapter());

		// replace all "%n" with channel name, "%d" with device name and all "%c" with chip name
		QStringList argList = watchCommand.split(" ",QString::SkipEmptyParts).replaceInStrings("%d", devName).replaceInStrings("%c", chip_name).replaceInStrings("%n", names[current_channel_idx]);
		if (argList.isEmpty())
			ui->textEdit->append("No watch app command set!");
		else {
			pWatcher = new QProcess();
			pWatcher->setProgram(argList.front());
			argList.removeFirst();
			pWatcher->setArguments(argList);
			pWatcher->start();
			connect(pWatcher, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(on_watchFinished(int, QProcess::ExitStatus)));
		}
	}

	ui->watchButton->setEnabled(false);
}

void MainWindow::on_watchFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
	//delete pWatcher;
	pWatcher = nullptr;
	if (pZapper && pZapper->zapping)  // should not be here if not true
		ui->watchButton->setEnabled(true);
} // on_watchFinished()

/*void MainWindow::on_lastFirst_toggled(bool checked)
{
    useLastFirst = checked;
}*/

void MainWindow::closeEvent (QCloseEvent *event)
{
	if (event->spontaneous()) {
		hide();
		on_actionStop_Monitoring_triggered();

		//QMessageBox::StandardButton resBtn = QMessageBox::question( this, "Really Quit?",
		// tr("Are you sure you want to exit the app?\n"), QMessageBox::No | QMessageBox::Yes, QMessageBox::Yes);
		//if (resBtn != QMessageBox::Yes)
			//event->ignore();
		//else
		//	event->accept();
	}
}

void MainWindow::saveRecentFilenames(void)
{
	QSettings settings(iniFname, QSettings::NativeFormat);
	QStringList files = settings.value("recentFileList").toStringList();

	files.removeAll(fileName);
	files.prepend(fileName);
	while (files.size() > numRecentFiles)
		files.removeLast();
	settings.setValue("recentFileList", files);
	updateRecentFileActions(settings);

} // setCurrentFile()

void MainWindow::updateRecentFileActions(QSettings & appSettings)
{
	QStringList files = appSettings.value("recentFileList").toStringList();

	for (int i = 0; i < files.size(); ++i) {
		//QString text = tr("&%1 %2").arg(i + 1).arg(strippedName(files[i]));
		recentFileActs[i]->setText(files[i]); //text);
		recentFileActs[i]->setData(files[i]);
		recentFileActs[i]->setVisible(true);
	}

	for (int j = files.size(); j < MAX_RECENT_FILES; ++j)
		recentFileActs[j]->setVisible(false);

} //updateRecentFileActions()

void MainWindow::on_chip_comboBox_currentIndexChanged(int index)
{
	if (index)
		strncpy(specific_chip, QS_to_char(ui->chip_comboBox->currentText()), 127);
	else
		memset(specific_chip, 0, 128);
}

void MainWindow::on_adapter_comboBox_currentIndexChanged(int index)
{
	if (index)
		specific_adapter = index - 1;
	else
		specific_adapter = -1;
}
