#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include "zap.h"
#include <QTimer>
#include <QRegExpValidator>
#include <QProcess>
#include <QCloseEvent>
#include <QSettings>

#define MAX_RECENT_FILES 20

QT_BEGIN_NAMESPACE
namespace Ui { class MainWindow; }
QT_END_NAMESPACE

class MainWindow : public QMainWindow
{
	Q_OBJECT

public:
	MainWindow(QWidget *parent = nullptr);
	~MainWindow();

private slots:
	void on_actionExit_triggered();
	void on_actionLoad_triggered();
	void on_channel_comboBox_currentIndexChanged(int index);
	void on_init_program();
	void showInfo(const char * txt); // receives signal sendInfo(const char* info);
	void updateStrength(float pct, unsigned int snr, bool has_lock); // receives signal newReading(float pct, bool has_lock);
	void zapperFinished();
	void saveSetting(QString settingName, QString settingValue, bool saveIt);
	void on_actionStop_Monitoring_triggered();
	void on_stopButton_clicked();
	void on_actionSettings_triggered();
	void on_channelTable_cellDoubleClicked(int row, int column);
	void on_actionScanSelected_triggered();
	void on_logCheckBox_clicked();
	void on_channelTable_itemSelectionChanged();
	void on_watchButton_clicked();
	void on_watchFinished(int exitCode, QProcess::ExitStatus exitStatus);
	void saveRecentFilenames(void);
	void updateRecentFileActions(QSettings & appSettings);
	void openRecentFile(void);
	
	//void on_lastFirst_toggled(bool checked);
	
	void on_chip_comboBox_currentIndexChanged(int index);
	
	void on_adapter_comboBox_currentIndexChanged(int index);
	
	private:
	void load_channels(void);
	void showEvent(QShowEvent *);
	void closeEvent(QCloseEvent *event);

	Ui::MainWindow *ui;

	QString fileName;
	QString newfile;
	QString iniFname;
	QString logFname;
	QString watchCommand;
	int current_channel_idx;
	Zapper * pZapper;
	QRegExpValidator* pValidator;
	QProcess * pWatcher;
	QTimer timer;
	float max;
	float min;
	float mean;
	int max_snr;
	int min_snr;
	int mean_snr;
	int num_rdgs;
	bool loading_file;
	bool save_filename;
	//bool was_watching;
	int lock_count;
	int numRecentFiles;
	QAction *recentFileActs[MAX_RECENT_FILES];

// Define signal:
signals:
	void sendWatchButtonClick();

};

#endif // MAINWINDOW_H

