#!/usr/bin/perl -w

use strict;
use POSIX;

die "Usage: merge_channels.pl files\n" if $#ARGV < 1;

my $date_fmt = '%b %e, %Y at %l:%M %p';

# KCTS-HD:189028615:8VSB:49:52:3

my %ctimes;
my %names;
my %apids;
my %vpids;
my %count;
my %freqs;
my %tps;

my $maxnamelen = 0;
my $maxname = '';

sub nameout
{
	my $spaces = 10 - length($_[0]);

	print ' ' x $spaces, $_[0];
}

foreach my $f (@ARGV) {

	open(INF, $f) || die "Could not open $f: $!\n";

	my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat(INF);

	CHANNEL:
	while (<INF>) {
		chomp;
		my @fields = split /:/;
		chomp @fields;

		die("$_ from $f split into ", scalar @fields, " fields\n") if scalar @fields != 6;
	
		#   0       1       2   3  4  5
		# WORLD:189028615:8VSB:97:100:6
		# name = 0, freq = 1, 8VSB = 2, apid = 3, vpid = 4, transport = 5

		die("Mod type = " . $fields[2] . " in $_ from $f\n") if $fields[2] ne "8VSB";

		my $freq = $fields[1];
		my $tp = $fields[5];

		die("$_ from $f has frequency of $freq\n") if ($freq =~ /\D/) || ($freq < 1000000) || ($freq > 999999999);
		die("$_ from $f has transport ID of $tp\n") if ($tp =~ /\D/) || ($tp < 1) || ($tp > 99);

		my $freq_tp = "$freq.$tp";

		my $NAME = uc $fields[0];
		$NAME =~ s/^\s+|\s+$//g;
		$NAME =~ s/\s/_/g;

		if (length($NAME) > 10) {
			my @words = split /_/, $NAME;
			$NAME = $words[0];
			#print "Truncated to \"$NAME\" in \"$_\"\n";
		}

		# if this name already defined, see which one is newer
		if (defined($ctimes{$freq_tp})) {
			#my $old_d = POSIX::strftime($date_fmt, localtime($ctimes{$freq_tp}));
			#my $new_d = POSIX::strftime($date_fmt, localtime($ctime));
			#$old_d =~ s/  / /g;
			#$new_d =~ s/  / /g;

			if ($ctime < $ctimes{$freq_tp}) { # this file is older, move on
				#print 'Old channel ';
				#nameout($names{$freq_tp});
				#print ' (found ', $old_d, ") is newer than ";
				#nameout($NAME);
				#print " (found ", $new_d, "), keeping old for $freq_tp\n";
				next CHANNEL;
			}

			# notify new callsign for this freq/transport channel
			#if ($names{$freq_tp} ne $NAME) {
			#	print 'Replacing ';
			#	nameout($names{$freq_tp});
			#	print ' (found ', $old_d, ') with ';
			#	nameout($NAME);
			#	print ' (found ', $new_d, ") for $freq_tp\n"
			#}
		}
		$ctimes{$freq_tp} = $ctime;

		# save name
		$names{$freq_tp} = $NAME;

		if (length($NAME) > $maxnamelen) {
			$maxnamelen = length($NAME);
			$maxname = $NAME;
		}

		# save the frequency and transport ID
		$freqs{$freq_tp} = $freq;
		$tps{$freq_tp} = $tp;

		# save PIDs
		$apids{$freq_tp} = $fields[3];
		$vpids{$freq_tp} = $fields[4];

		# how many times did this station show up?
		$count{$freq_tp}++;
	}
	close INF;
}

open(OUTF, ">allchans.conf") || die "Could not open allchans.conf for output: $!\n";

# frequency.transport uniquely identifies a station
my @freq_tps = sort keys %count;
print scalar @freq_tps, " stations found\n";

my %namecount;

foreach my $ftp (@freq_tps) {
	while (defined($namecount{$names{$ftp}})) {
		if ($names{$ftp} =~ /(\d)$/) {
			my $next_digit = $1 + 1;
			$names{$ftp} =~ s/(\d)$/$next_digit/;
		}
		else {
			$names{$ftp} .= '-1';
		}
	}

	$namecount{$names{$ftp}}++;
	my $fd = POSIX::strftime($date_fmt, localtime($ctimes{$ftp}));

	print "$freqs{$ftp}  tp $tps{$ftp}:  ";
	nameout($names{$ftp});
	print " found $fd\n";

	# WORLD:189028615:8VSB:97:100:6
	print OUTF "$names{$ftp}:$freqs{$ftp}:8VSB:$apids{$ftp}:$vpids{$ftp}:$tps{$ftp}\n";
}

close OUTF;

#print "Max name was $maxname, length = $maxnamelen\n";

